/*
 * CPM (clock and power management) support for the X1600.
 *
 * Copyright (C) 2017, 2018, 2020, 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/devices/cpm.h>

#include <l4/sys/types.h>
#include <stdint.h>



#ifdef __cplusplus

#include <l4/devices/hw_register_block.h>

/* Register access type. */

class Cpm_regs
{
  Hw::Register_block<32> _regs;

public:
  uint32_t exclk_freq;

  explicit Cpm_regs(l4_addr_t addr, uint32_t exclk_freq);

  // Utility methods.

  uint32_t get_field(uint32_t reg, uint32_t mask, uint8_t shift);
  void set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value);
};

/* A simple abstraction for accessing the CPM registers.
 * A proper device could inherit from Hw::Device and use an
 * Int_property for _exclk_freq. */

class Cpm_x1600_chip : public Cpm_chip
{
private:
  Hw::Register_block<32> _regs;
  uint32_t _exclk_freq;
  Cpm_regs _cpm_regs;

public:
  Cpm_x1600_chip(l4_addr_t addr, uint32_t exclk_freq);

  int have_clock(enum Clock_identifiers clock);
  void start_clock(enum Clock_identifiers clock);
  void stop_clock(enum Clock_identifiers clock);

  // Clock dividers.

  uint32_t get_divider(enum Clock_identifiers clock);
  void set_divider(enum Clock_identifiers clock, uint32_t divider);

  // Clock sources.

  uint8_t get_source(enum Clock_identifiers clock);
  void set_source(enum Clock_identifiers clock, uint8_t source);

  // Source frequencies.

  uint32_t get_source_frequency(enum Clock_identifiers clock);

  // Output clock frequencies.

  uint32_t get_frequency(enum Clock_identifiers clock);
  void set_frequency(enum Clock_identifiers clock, uint32_t frequency);

  // Other operations.

  void set_pll_parameters(enum Clock_identifiers clock, uint16_t multiplier,
                          uint8_t in_divider, uint8_t out_divider);
};

#endif /* __cplusplus */



/* C language interface. */

EXTERN_C_BEGIN

void *x1600_cpm_init(l4_addr_t cpm_base);

int x1600_cpm_have_clock(void *cpm, enum Clock_identifiers clock);
void x1600_cpm_start_clock(void *cpm, enum Clock_identifiers clock);
void x1600_cpm_stop_clock(void *cpm, enum Clock_identifiers clock);

uint32_t x1600_cpm_get_divider(void *cpm, enum Clock_identifiers clock);
void x1600_cpm_set_divider(void *cpm, enum Clock_identifiers clock, uint32_t divider);

uint8_t x1600_cpm_get_source(void *cpm, enum Clock_identifiers clock);
void x1600_cpm_set_source(void *cpm, enum Clock_identifiers clock, uint8_t source);

uint32_t x1600_cpm_get_source_frequency(void *cpm, enum Clock_identifiers clock);

uint32_t x1600_cpm_get_frequency(void *cpm, enum Clock_identifiers clock);
void x1600_cpm_set_frequency(void *cpm, enum Clock_identifiers clock, uint32_t frequency);

void x1600_cpm_set_mpll_parameters(void *cpm, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider);

EXTERN_C_END
